# Copyright © 2021 Chris Hughes
########################################################################
# This example trains a ResNet50-D on the Imagenette Dataset using components from the torchvision, timm and
# torchmetrics libraries.
# This example demonstrates how the trainer can be extended to incorporate techniques such as mixup and modelEMA
# into a training run.
#
# Note: this example requires installing the torchvision, torchmetrics and timm packages
########################################################################

from pathlib import Path

import timm
import timm.data
import timm.loss
import timm.optim
import timm.utils
import timm.scheduler
import torch
import torchmetrics
from func_to_script import script

from pytorch_accelerated.callbacks import SaveBestModelCallback
from pytorch_accelerated.trainer import DEFAULT_CALLBACKS, Trainer


def create_datasets(image_size, data_mean, data_std, train_path, val_path):
    train_transforms = timm.data.create_transform(
        input_size=image_size,
        is_training=True,
        mean=data_mean,
        std=data_std,
        auto_augment="rand-m7-mstd0.5-inc1",
    )

    eval_transforms = timm.data.create_transform(
        input_size=image_size, mean=data_mean, std=data_std
    )

    train_dataset = timm.data.dataset.ImageDataset(
        train_path, transform=train_transforms
    )
    eval_dataset = timm.data.dataset.ImageDataset(val_path, transform=eval_transforms)

    return train_dataset, eval_dataset


class TimmMixupTrainer(Trainer):
    def __init__(self, eval_loss_fn, mixup_args, num_classes, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.eval_loss_fn = eval_loss_fn
        self.num_updates = None
        self.mixup_fn = timm.data.Mixup(**mixup_args)

        self.accuracy = torchmetrics.Accuracy(
            task="multiclass", num_classes=num_classes
        )
        self.ema_accuracy = torchmetrics.Accuracy(
            task="multiclass", num_classes=num_classes
        )
        self.ema_model = None

    def create_scheduler(self):
        return timm.scheduler.CosineLRScheduler(
            self.optimizer,
            t_initial=self.run_config.num_epochs,
            cycle_decay=0.5,
            lr_min=1e-6,
            t_in_epochs=True,
            warmup_t=3,
            warmup_lr_init=1e-4,
            cycle_limit=1,
        )

    def training_run_start(self):
        # Model EMA requires the model without a DDP wrapper and before sync batchnorm conversion
        self.ema_model = timm.utils.ModelEmaV2(
            self._accelerator.unwrap_model(self.model), decay=0.9
        )
        if self.run_config.is_distributed:
            self.model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(self.model)

    def train_epoch_start(self):
        super().train_epoch_start()
        self.num_updates = self.run_history.current_epoch * len(self._train_dataloader)

    def calculate_train_batch_loss(self, batch):
        xb, yb = batch
        mixup_xb, mixup_yb = self.mixup_fn(xb, yb)
        return super().calculate_train_batch_loss((mixup_xb, mixup_yb))

    def train_epoch_end(
        self,
    ):
        self.ema_model.update(self.model)
        self.ema_model.eval()

        if hasattr(self.optimizer, "sync_lookahead"):
            self.optimizer.sync_lookahead()

    def scheduler_step(self):
        self.num_updates += 1
        if self.scheduler is not None:
            self.scheduler.step_update(num_updates=self.num_updates)

    def calculate_eval_batch_loss(self, batch):
        with torch.no_grad():
            xb, yb = batch
            outputs = self.model(xb)
            val_loss = self.eval_loss_fn(outputs, yb)
            self.accuracy.update(outputs.argmax(-1), yb)

            ema_model_preds = self.ema_model.module(xb).argmax(-1)
            self.ema_accuracy.update(ema_model_preds, yb)

        return {"loss": val_loss, "model_outputs": outputs, "batch_size": xb.size(0)}

    def eval_epoch_end(self):
        super().eval_epoch_end()

        if self.scheduler is not None:
            self.scheduler.step(self.run_history.current_epoch + 1)

        self.run_history.update_metric("accuracy", self.accuracy.compute().cpu())
        self.run_history.update_metric(
            "ema_model_accuracy", self.ema_accuracy.compute().cpu()
        )
        self.accuracy.reset()
        self.ema_accuracy.reset()


DATA_PATH = (
    Path("/".join(Path(__file__).absolute().parts[:-3])) / "data/imagenette2-320"
)


@script
def main(
    data_path: str = DATA_PATH,
    image_size: int = 224,
    lr: float = 5e-3,
    batch_size: int = 32,
    num_epochs: int = 40,
    pretrained: bool = False,
    smoothing: float = 0.1,
    mixup: float = 0.2,
    cutmix: float = 1.0,
    bce_target_thresh: float = 0.2,
):
    data_path = Path(data_path)
    train_path = data_path / "train"
    val_path = data_path / "val"
    num_classes = len(list(train_path.iterdir()))
    image_size = (image_size, image_size)

    mixup_args = dict(
        mixup_alpha=mixup,
        cutmix_alpha=cutmix,
        label_smoothing=smoothing,
        num_classes=num_classes,
    )

    # Create model using timm
    model = timm.create_model(
        "resnet50d", pretrained=pretrained, num_classes=num_classes, drop_path_rate=0.05
    )

    # Load data config associated with the model to use in data augmentation pipeline
    data_config = timm.data.resolve_data_config({}, model=model, verbose=True)
    data_mean = data_config["mean"]
    data_std = data_config["std"]

    # Create training and validation datasets
    train_dataset, eval_dataset = create_datasets(
        train_path=train_path,
        val_path=val_path,
        image_size=image_size,
        data_mean=data_mean,
        data_std=data_std,
    )

    # Create optimizer
    optimizer = timm.optim.create_optimizer_v2(
        model, opt="lookahead_AdamW", lr=lr, weight_decay=0.01
    )

    # As we are using Mixup, we can use BCE during training and CE for evaluation
    train_loss_fn = timm.loss.BinaryCrossEntropy(
        target_threshold=bce_target_thresh, smoothing=smoothing
    )
    validate_loss_fn = torch.nn.CrossEntropyLoss()

    # Create trainer and start training
    trainer = TimmMixupTrainer(
        model=model,
        optimizer=optimizer,
        loss_func=train_loss_fn,
        eval_loss_fn=validate_loss_fn,
        mixup_args=mixup_args,
        num_classes=num_classes,
        callbacks=[
            *DEFAULT_CALLBACKS,
            SaveBestModelCallback(watch_metric="accuracy", greater_is_better=True),
        ],
    )

    trainer.train(
        per_device_batch_size=batch_size,
        train_dataset=train_dataset,
        train_dataloader_kwargs={"drop_last": True},  # mixup requires even batches
        eval_dataset=eval_dataset,
        num_epochs=num_epochs,
        create_scheduler_fn=trainer.create_scheduler,
    )


if __name__ == "__main__":
    main()



# Modifications Copyright © 2022 Chris Hughes
# Adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/scheduler/cosine_lr.py

import math
from functools import partial
from typing import Callable

import torch

from pytorch_accelerated import TrainerPlaceholderValues
from pytorch_accelerated.schedulers.scheduler_base import StatefulSchedulerBase


class CosineLrScheduler(StatefulSchedulerBase):
    """
    A stateful Cosine annealing learning rate scheduler, as described in `this paper <https://arxiv.org/abs/1608.03983>`_,
    but without restarts.

    This scheduler differs from the PyTorch's :class:`~torch.optim.lr_scheduler.CosineAnnealingLR` as it provides options
    to add warmup and cooldown epochs. Additionally, the annealing rate can be modified by adjusting the k-decay
    parameter, for which the rate of change of the learning rate is changed by its k-th order derivative,
    as described in `here <https://arxiv.org/abs/2004.05909>`_.

    If warmup epochs are specified, the learning rate will increase in constant increments from the ``warmup_starting_lr``
    provided until the learning rate specified in the parameter group is reached.

    If cooldown epochs are specified, the learning rate will be fixed at the minimum lr value given. This behaviour will
    continue if the scheduler is called after the training cycle has completed.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        total_num_epochs: int,
        num_update_steps_per_epoch: int,
        k_decay=1.0,
        lr_min: float = 1e-6,
        min_lr_ratio=None,
        num_warmup_epochs: int = 0,
        warmup_starting_lr=1e-6,
        warmup_starting_lr_ratio=None,
        num_cooldown_epochs=0,
    ):
        """
        Create a new ConsineLrScheduler object which can be used to modify the learning rate in an optimizer's parameter
        groups.

        :param optimizer: a PyTorch optimizer containing one or more parameter groups
        :param total_num_epochs: the total number of training epochs, inclusive of any warmup and cooldown epochs
        :param num_update_steps_per_epoch: the number of optimizer updates that take place per epoch
        :param k_decay: adjusts the rate of annealing. Higher values will maintain a higher lr for longer
        :param lr_min: the minimum value that the learning rate should decay to for all parameter groups. This will be held fixed during cooldown epochs
        :param min_lr_ratio: this can be used to represent the minimum lr for each parameter group as a ratio of its maximum lr. If set, this will take precedence over ``lr_min``
        :param num_warmup_epochs: the number of epochs to gradually increase the lr until it reaches the maximum value
        :param warmup_starting_lr: the starting lr that will be used for all parameter groups at the beginning of training if ``num_warmup_epochs`` is greater than 0
        :param warmup_starting_lr_ratio: this can be used to represent the warmup starting lr for each parameter group as a ratio of its maximum lr. If set, this will take precedence over ``warmup_starting_lr``
        :param num_cooldown_epochs: the number of epochs to hold the lr at its minimum value
        """

        super().__init__(optimizer)
        assert total_num_epochs > 0 and num_update_steps_per_epoch > 0
        assert lr_min >= 0
        self.total_iterations = total_num_epochs * num_update_steps_per_epoch
        self.lr_min_ratio = min_lr_ratio
        self.lr_min = lr_min
        self.warmup_iterations = num_warmup_epochs * num_update_steps_per_epoch
        self.warmup_lr_init = warmup_starting_lr
        self.warmup_lr_ratio = warmup_starting_lr_ratio
        self.k_decay = k_decay
        self.num_cooldown_iterations = num_cooldown_epochs * num_update_steps_per_epoch
        if self.warmup_iterations:
            self._update_param_groups(self.warmup_lr_init)

    def get_updated_values(self, num_updates: int):
        """
        Calculate the learning rate for a particular step given the number of previous updates.

        If warmup epochs are specified, the learning rate will increase in constant increments from the ``warmup_starting_lr``
        provided until the learning rate specified in the parameter group is reached.

        If cooldown epochs are specified, the learning rate will be fixed at the minimum lr value given. This behaviour will
        continue if the scheduler is called after the training cycle has completed.

        Between any warmup or cooldown epochs, the cosine annealing strategy will be used.

        :param num_updates: the number of previous updates
        :return: the learning rates with which to update each parameter group
        """
        if num_updates < self.warmup_iterations:
            # increase lr linearly
            lrs = [
                (
                    self.warmup_lr_ratio * lr
                    if self.warmup_lr_ratio is not None
                    else self.warmup_lr_init
                )
                + num_updates * ((lr - self.warmup_lr_init) / self.warmup_iterations)
                for lr in self.base_lr_values
            ]
        elif num_updates < (
            self.total_iterations
            - (self.warmup_iterations + self.num_cooldown_iterations)
        ):
            num_updates = num_updates - self.warmup_iterations
            total_cosine_iterations = self.total_iterations - (
                self.warmup_iterations + self.num_cooldown_iterations
            )

            lrs = [
                (
                    self.lr_min_ratio * lr_max
                    if self.lr_min_ratio is not None
                    else self.lr_min
                )
                + 0.5
                * (
                    lr_max
                    - (
                        self.lr_min_ratio * lr_max
                        if self.lr_min_ratio is not None
                        else self.lr_min
                    )
                )
                * (
                    1
                    + math.cos(
                        math.pi
                        * num_updates**self.k_decay
                        / total_cosine_iterations**self.k_decay
                    )
                )
                for lr_max in self.base_lr_values
            ]

        else:
            # cooldown
            lrs = [
                self.lr_min_ratio * base_lr
                if self.lr_min_ratio is not None
                else self.lr_min
                for base_lr in self.base_lr_values
            ]

        return lrs

    @classmethod
    def create_scheduler_fn(
        cls,
        total_num_epochs: int = TrainerPlaceholderValues.NUM_EPOCHS,
        num_update_steps_per_epoch: int = TrainerPlaceholderValues.NUM_UPDATE_STEPS_PER_EPOCH,
        k_decay=1.0,
        lr_min: float = 1e-6,
        min_lr_ratio=None,
        num_warmup_epochs: int = 0,
        warmup_starting_lr=1e-6,
        warmup_starting_lr_ratio=None,
        num_cooldown_epochs=0,
    ) -> Callable:
        """
        An alternative constructor which returns a function that accepts an optimizer and creates an instance of
        ``CosineLrScheduler``. This is primarily intended to be used with the :class:`~pytorch_accelerated.trainer.Trainer`
        as illustrated below::

            trainer.train(
            train_dataset=train_dataset,
            num_epochs=num_epochs,
            per_device_batch_size=batch_size,
            create_scheduler_fn=CosineLrScheduler.create_scheduler_fn(num_warmup_epochs=5),
            )

        By default, the ``total_num_epochs`` and ``num_iterations_per_epoch`` arguments will be set by the
        :class:`~pytorch_accelerated.trainer.Trainer` with the correct values at runtime.

        :param total_num_epochs: the total number of training epochs, inclusive of any warmup and cooldown epochs
        :param num_update_steps_per_epoch: the number of optimizer updates that take place per epoch
        :param k_decay: adjusts the rate of annealing. Higher values will maintain a higher lr for longer
        :param lr_min: the minimum value that the learning rate should decay to for all parameter groups. This will be held fixed during cooldown epochs
        :param min_lr_ratio: this can be used to represent the minimum lr for each parameter group as a ratio of its maximum lr. If set, this will take precedence over ``lr_min``
        :param num_warmup_epochs: the number of epochs to gradually increase the lr until it reaches the maximum value
        :param warmup_starting_lr: the starting lr that will be used for all parameter groups at the beginning of training if ``num_warmup_epochs`` is greater than 0
        :param warmup_starting_lr_ratio: this can be used to represent the warmup starting lr for each parameter group as a ratio of its maximum lr. If set, this will take precedence over ``warmup_starting_lr``
        :param num_cooldown_epochs: the number of epochs to hold the lr at its minimum value
        :return: a function which accepts an optimizer as an argument and returns an instance of :class:`CosineLrScheduler`
        """
        return partial(
            cls,
            total_num_epochs=total_num_epochs,
            num_update_steps_per_epoch=num_update_steps_per_epoch,
            lr_min=lr_min,
            min_lr_ratio=min_lr_ratio,
            k_decay=k_decay,
            num_warmup_epochs=num_warmup_epochs,
            warmup_starting_lr=warmup_starting_lr,
            warmup_starting_lr_ratio=warmup_starting_lr_ratio,
            num_cooldown_epochs=num_cooldown_epochs,
        )



# Modifications Copyright © 2021 Chris Hughes
########################################################################
# This is an accelerated example of the PyTorch "Transfer Learning for Computer Vision Tutorial"
# written by Sasank Chilamkurthy, available here:
# https://pytorch.org/tutorials/beginner/transfer_learning_tutorial.html,
# which demonstrates fine-tuning a ResNet18 model to classify ants and bees.
#
# Note: this example requires installing the torchvision package
########################################################################
import os
from functools import partial
from pathlib import Path

from func_to_script import script
from torch import nn, optim
from torch.optim import lr_scheduler
from torchvision import datasets, models, transforms
from torchvision.models import ResNet18_Weights

from pytorch_accelerated.trainer import Trainer, TrainerPlaceholderValues

DATA_PATH = (
    Path("/".join(Path(__file__).absolute().parts[:-3])) / "data/hymenoptera_data"
)


@script
def main(data_dir: str = DATA_PATH):
    # Data augmentation and normalization for training
    # Just normalization for validation
    data_transforms = {
        "train": transforms.Compose(
            [
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
            ]
        ),
        "val": transforms.Compose(
            [
                transforms.Resize(256),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
            ]
        ),
    }

    # Create datasets
    image_datasets = {
        x: datasets.ImageFolder(os.path.join(data_dir, x), data_transforms[x])
        for x in ["train", "val"]
    }
    # Create model
    model = models.resnet18(weights=ResNet18_Weights.DEFAULT)
    model.fc = nn.Linear(model.fc.in_features, len(image_datasets["train"].classes))

    # Define loss function
    loss_func = nn.CrossEntropyLoss()

    # Create optimizer and scheduler
    optimizer = optim.SGD(model.parameters(), lr=0.001, momentum=0.9)

    # The Trainer calls schedulers after every step, not every epoch as StepLr expects
    # To use this as intended, we need to represent the step size as the number of iterations
    exp_lr_scheduler = partial(
        lr_scheduler.StepLR,
        step_size=TrainerPlaceholderValues.NUM_UPDATE_STEPS_PER_EPOCH * 7,
        gamma=0.1,
    )

    trainer = Trainer(
        model,
        loss_func=loss_func,
        optimizer=optimizer,
    )

    trainer.train(
        train_dataset=image_datasets["train"],
        eval_dataset=image_datasets["val"],
        num_epochs=8,
        per_device_batch_size=4,
        create_scheduler_fn=exp_lr_scheduler,
    )


if __name__ == "__main__":
    main()



import math
from functools import partial

from pytest import mark

from pytorch_accelerated import TrainerPlaceholderValues
from pytorch_accelerated.run_config import TrainerRunConfig
from pytorch_accelerated.trainer import replace_trainer_placeholder_values


def create_run_config(
    train_dl_length,
    num_epochs=5,
    gradient_accumulation_steps=1,
    num_processes=1,
    train_batch_size=32,
    eval_batch_size=64,
):
    num_update_steps_per_epoch = math.ceil(
        train_dl_length / gradient_accumulation_steps
    )

    max_num_train_steps = num_epochs * num_update_steps_per_epoch

    return TrainerRunConfig(
        num_epochs=num_epochs,
        train_per_device_batch_size=train_batch_size,
        train_dl_kwargs={"tkwarg1": "tval1"},
        eval_per_device_batch_size=eval_batch_size,
        eval_dl_kwargs={"ekwarg1": "eval1"},
        gradient_accumulation_steps=gradient_accumulation_steps,
        gradient_clip_value=None,
        train_total_batch_size=train_batch_size
        * num_processes
        * gradient_accumulation_steps,
        eval_total_batch_size=eval_batch_size * num_processes,
        max_num_train_steps=max_num_train_steps,
        is_local_process_zero=True,
        is_world_process_zero=True,
        is_distributed=True,
        mixed_precision="fp16",
        num_processes=1,
        num_update_steps_per_epoch=num_update_steps_per_epoch,
    )


class FakeDataLoader:
    def __init__(self, length):
        self.length = length

    def __len__(self):
        return self.length


class FakeTrainer:
    def __init__(
        self,
        train_dl_length,
        eval_dl_length,
        run_config,
    ):
        self._train_dataloader = FakeDataLoader(train_dl_length)
        self._eval_dataloader = FakeDataLoader(eval_dl_length)
        self.run_config = run_config


@mark.parametrize(
    (
        "expected_train_dl_length",
        "expected_eval_dl_length",
        "num_epochs",
        "num_processes",
        "gradient_accumulation_steps",
    ),
    [(1000, 500, 5, 1, 1), (800, 200, 10, 2, 2)],
)
def test_can_inject_placeholders(
    expected_train_dl_length,
    expected_eval_dl_length,
    num_epochs,
    num_processes,
    gradient_accumulation_steps,
):
    int_arg = 3
    str_arg = "test"

    trainer = FakeTrainer(
        expected_train_dl_length,
        expected_eval_dl_length,
        create_run_config(
            num_epochs=num_epochs,
            num_processes=num_processes,
            gradient_accumulation_steps=gradient_accumulation_steps,
            train_dl_length=expected_train_dl_length,
        ),
    )
    fn_with_placeholders = partial(
        dict,
        num_update_steps=TrainerPlaceholderValues.NUM_UPDATE_STEPS_PER_EPOCH,
        train_dl_length=TrainerPlaceholderValues.TRAIN_DATALOADER_LEN,
        eval_dl_length=TrainerPlaceholderValues.EVAL_DATALOADER_LEN,
        num_epochs=TrainerPlaceholderValues.NUM_EPOCHS,
        int_arg=int_arg,
        str_arg=str_arg,
    )

    replaced_fn = replace_trainer_placeholder_values(trainer, fn_with_placeholders)
    keyword_dict = replaced_fn()

    assert (
        keyword_dict["num_update_steps"]
        == trainer.run_config.num_update_steps_per_epoch
    )
    assert keyword_dict["num_epochs"] == num_epochs
    assert keyword_dict["train_dl_length"] == expected_train_dl_length
    assert keyword_dict["eval_dl_length"] == expected_eval_dl_length
    assert keyword_dict["int_arg"] == int_arg
    assert keyword_dict["str_arg"] == str_arg


def test_can_multiply_placeholder():
    expected_train_dl_length = 100
    expected_eval_dl_length = 50
    num_epochs = 5
    trainer = FakeTrainer(
        expected_train_dl_length,
        expected_eval_dl_length,
        create_run_config(
            num_epochs=num_epochs,
            train_dl_length=expected_train_dl_length,
        ),
    )
    fn_with_placeholders = partial(
        dict,
        num_epochs=TrainerPlaceholderValues.NUM_EPOCHS,
        num_epochs_modified=TrainerPlaceholderValues.NUM_EPOCHS * 2,
    )

    replaced_fn = replace_trainer_placeholder_values(trainer, fn_with_placeholders)
    keyword_dict = replaced_fn()

    assert keyword_dict["num_epochs"] == num_epochs
    assert keyword_dict["num_epochs_modified"] == 2 * num_epochs


def test_can_add_placeholder():
    expected_train_dl_length = 100
    expected_eval_dl_length = 50
    num_epochs = 5
    trainer = FakeTrainer(
        expected_train_dl_length,
        expected_eval_dl_length,
        create_run_config(
            num_epochs=num_epochs,
            train_dl_length=expected_train_dl_length,
        ),
    )
    fn_with_placeholders = partial(
        dict,
        num_epochs=TrainerPlaceholderValues.NUM_EPOCHS,
        num_epochs_modified=TrainerPlaceholderValues.NUM_EPOCHS + 2,
    )

    replaced_fn = replace_trainer_placeholder_values(trainer, fn_with_placeholders)
    keyword_dict = replaced_fn()

    assert keyword_dict["num_epochs"] == num_epochs
    assert keyword_dict["num_epochs_modified"] == 2 + num_epochs



from pytest import fixture, raises

from pytorch_accelerated.tracking import InMemoryRunHistory


@fixture
def run_history() -> InMemoryRunHistory:
    return InMemoryRunHistory()


def test_can_increment_epoch(run_history: InMemoryRunHistory):
    initial_epoch_count = run_history.current_epoch

    run_history._increment_epoch()
    incremented_once = run_history.current_epoch
    run_history._increment_epoch()
    incremented_twice = run_history.current_epoch

    assert incremented_once == initial_epoch_count + 1
    assert incremented_twice == initial_epoch_count + 2


def test_can_get_metric_values(run_history: InMemoryRunHistory):
    metric_name = "my_metric"
    expected_metric_values = [0, 1, 2]
    run_history._metrics[metric_name].extend(expected_metric_values)

    metric_values = run_history.get_metric_values(metric_name)

    assert metric_values == expected_metric_values


def test_can_update_metric(run_history: InMemoryRunHistory):
    metric_name = "my_metric"
    expected_metric_values = [0, 1, 2]

    for value in expected_metric_values:
        run_history.update_metric(metric_name, value)

    assert run_history.get_metric_values(metric_name) == expected_metric_values


def test_can_get_latest_metric(run_history: InMemoryRunHistory):
    metric_name = "my_metric"
    metric_values = [0, 1, 2]
    for value in metric_values:
        run_history.update_metric(metric_name, value)

    latest_metric = run_history.get_latest_metric(metric_name)

    assert latest_metric == metric_values[-1]


def test_can_not_get_latest_metric_if_not_recorded(run_history: InMemoryRunHistory):
    metric_name = "my_metric"
    with raises(ValueError):
        run_history.get_latest_metric(metric_name)


def test_can_get_metric_names(run_history: InMemoryRunHistory):
    expected_metric_names = {"my_metric_1", "my_metric_2"}
    expected_metric_values = [0, 1]
    for name, value in zip(expected_metric_names, expected_metric_values):
        run_history.update_metric(name, value)

    metric_names = run_history.get_metric_names()

    assert metric_names == expected_metric_names


def test_can_reset(run_history: InMemoryRunHistory):
    metric_names = {"my_metric_1", "my_metric_2"}
    metric_values = [0, 1]
    for name, value in zip(metric_names, metric_values):
        run_history.update_metric(name, value)

    run_history.reset()

    assert len(run_history.get_metric_names()) == 0



from pytest import fixture, mark
from torch import nn

from pytorch_accelerated.finetuning import ModelFreezer, _module_is_batch_norm


class TestModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.input = nn.Linear(100, 100)
        self.block_1 = nn.Sequential(
            nn.Linear(100, 100),
            nn.BatchNorm1d(100),
            nn.ReLU(),
        )
        self.block_2 = nn.Sequential(
            nn.Linear(100, 100),
            nn.BatchNorm1d(100),
            nn.Sequential(
                nn.Linear(100, 100),
                nn.BatchNorm1d(100),
                nn.ReLU(),
            ),
        )
        self.output_1 = nn.Linear(100, 10)
        self.output_2 = nn.Linear(100, 10)

    def forward(self, x):
        x = self.input(x)
        x = self.block_1(x)
        x = self.block_2(x)
        out_1 = self.output_1(x)
        out_2 = self.output_2(x)
        return (out_1, out_2)


@fixture
def model():
    return TestModel()


def verify_layer_state(layer, frozen=True):
    params = list(layer.module.parameters(recurse=False))
    if params:
        requires_grad = {param.requires_grad for param in params}
        all_params_same_state = len(requires_grad) == 1
        all_params_correct_state = (not frozen) is list(requires_grad)[0]
        layer_state_correct = layer.is_frozen is frozen
    else:
        all_params_same_state = True
        all_params_correct_state = True
        if _module_is_batch_norm(layer.module):
            layer_state_correct = layer.is_frozen is frozen
        else:
            layer_state_correct = True

    return all_params_same_state and all_params_correct_state and layer_state_correct


@mark.parametrize(
    ["idxs", "expected_idxs"],
    [
        [(0, 4), (0, 4)],  # leaves idxs in correct format unchanged
        [(4, 0), (0, 4)],  # from_idx greater than to_idx
        [(-2, -1), (3, 4)],  # using neg idxs
        [(-1, -2), (3, 4)],  # reversed neg idxs
        [(0, -1), (0, 4)],  # using combination of pos, neg idxs
    ],
)
def test_can_convert_idxs(model, idxs, expected_idxs):
    freezer = ModelFreezer(model, freeze_batch_norms=True)

    actual_idxs = freezer._convert_idxs(*idxs)

    assert actual_idxs == expected_idxs


@mark.parametrize(
    ["from_index", "to_index"],
    [
        [0, 2],
        [0, 4],
        [1, 3],
    ],
)
def test_can_freeze_model(model, from_index, to_index):
    freezer = ModelFreezer(model, freeze_batch_norms=True)

    freezer.freeze(from_index=from_index, to_index=to_index)
    layers = freezer.get_layers()

    for layer in layers:
        if from_index <= layer.layer_group_idx[0] <= to_index:
            assert verify_layer_state(layer, frozen=True)
        else:
            assert verify_layer_state(layer, frozen=False)


@mark.parametrize(
    ["from_index", "to_index"],
    [
        [0, 2],
        [0, 4],
        [1, 3],
    ],
)
def test_can_freeze_model_except_batch_norm(model, from_index, to_index):
    freezer = ModelFreezer(model, freeze_batch_norms=False)

    freezer.freeze(from_index=from_index, to_index=to_index)
    layers = freezer.get_layers()

    for layer in layers:
        if from_index <= layer.layer_group_idx[0] <= to_index:
            if _module_is_batch_norm(layer.module):
                assert verify_layer_state(layer, frozen=False)
            else:
                assert verify_layer_state(layer, frozen=True)
        else:
            assert verify_layer_state(layer, frozen=False)


@mark.parametrize(
    ["from_index", "to_index"],
    [
        [0, 2],
        [0, 4],
        [1, 3],
    ],
)
def test_can_unfreeze_model(model, from_index, to_index):
    freezer = ModelFreezer(model, freeze_batch_norms=True)
    freezer.freeze(from_index=0, to_index=-1)

    freezer.unfreeze(from_index, to_index)
    layers = freezer.get_layers()

    for layer in layers:
        if from_index <= layer.layer_group_idx[0] <= to_index:
            assert verify_layer_state(layer, frozen=False)
        else:
            assert verify_layer_state(layer, frozen=True)


@mark.parametrize(
    ["from_index", "to_index"],
    [
        [0, 2],
        [0, 4],
        [1, 3],
    ],
)
def test_unfreeze_model_returns_correct_parameters_for_frozen_model(
    model, from_index, to_index
):
    freezer = ModelFreezer(model, freeze_batch_norms=True)
    freezer.freeze(from_index=0, to_index=-1)
    expected_params = {
        lg.layer_group_idx[0]: {"params": list(lg.module.parameters())}
        for lg in freezer.get_layer_groups()
        if from_index <= lg.layer_group_idx[0] <= to_index
    }

    actual_params = freezer.unfreeze(from_index, to_index)

    assert actual_params == expected_params


@mark.parametrize(
    ["from_index", "to_index"],
    [
        [0, 4],
    ],
)
def test_unfreeze_model_returns_correct_parameters_for_partially_frozen_model(
    model, from_index, to_index
):
    freezer = ModelFreezer(model, freeze_batch_norms=True)
    freezer.freeze(from_index=0, to_index=-2)
    expected_params = {
        lg.layer_group_idx[0]: {"params": list(lg.module.parameters())}
        for lg in freezer.get_layer_groups()
        if (from_index <= lg.layer_group_idx[0] <= to_index) and lg.is_frozen is True
    }

    actual_params = freezer.unfreeze(from_index, to_index)

    assert actual_params == expected_params



import torch
from pytorch_accelerated.tracking import LossTracker


def test_can_track_loss_incrementally():
    tracker = LossTracker()
    losses = [105.5, 94.7, 98.0, 55.2, 32.5]
    expected_loss = sum(losses) / len(losses)

    for loss in losses:
        tracker.update(loss)
    calculated_loss = tracker.average

    assert expected_loss == calculated_loss


def test_can_track_loss_for_varying_batch_sizes():
    tracker = LossTracker()
    losses = [105.5, 94.7, 98.0, 55.2, 32.5]
    batch_sizes = [64, 32, 16, 8, 4]
    total_count = sum(batch_sizes)
    total_loss = (torch.tensor(losses) * torch.tensor(batch_sizes)).sum().item()
    expected_loss = total_loss / total_count

    for loss, batch_size in zip(losses, batch_sizes):
        tracker.update(loss, batch_size=batch_size)
    calculated_loss = tracker.average

    assert expected_loss == calculated_loss


def test_calculate_loss_average_calculation():
    tracker = LossTracker()
    losses = [105.5, 94.7, 98.0, 55.2, 32.5]
    batch_sizes = [64, 32, 16, 8, 4]
    total_count = sum(batch_sizes)
    total_loss = (torch.tensor(losses) * torch.tensor(batch_sizes)).sum().item()
    expected_loss = total_loss / total_count

    for loss, batch_size in zip(losses, batch_sizes):
        tracker.update(loss, batch_size=batch_size)
    calculated_loss = tracker.total_loss / tracker.running_count

    assert expected_loss == calculated_loss



from unittest.mock import Mock
import pytest

import torch
from pytorch_accelerated import Trainer
from pytorch_accelerated.callbacks import LimitBatchesCallback
from pytorch_accelerated.utils import (
    LimitBatches,
    local_process_zero_first,
    local_process_zero_only,
    remove_padding,
    world_process_zero_only,
)

PAD_VALUE = -1


def test_can_limit_batches_with_manager():
    limit_batches_num = 4
    with LimitBatches(limit_batches_num):
        limited_trainer = Trainer(model=Mock(), loss_func=Mock(), optimizer=Mock())
    callback = limited_trainer.callback_handler.callbacks[0]

    assert isinstance(callback, LimitBatchesCallback)
    assert callback.num_batches == limit_batches_num


def test_limit_batches_not_present_outside_manager():
    with LimitBatches(2):
        pass

    trainer = Trainer(model=Mock(), loss_func=Mock(), optimizer=Mock())

    assert LimitBatchesCallback not in {
        callback.__class__ for callback in trainer.callback_handler.callbacks
    }


def test_can_remove_padding_1d():
    t = torch.tensor([0, 1, 2, 3, PAD_VALUE, PAD_VALUE])
    expected_t = torch.tensor([0, 1, 2, 3])

    actual_t = remove_padding(t, PAD_VALUE)

    assert torch.eq(expected_t, actual_t).all()


def test_can_remove_padding_2d():
    t = torch.tensor([[0, 1], [1, 2], [PAD_VALUE, PAD_VALUE], [PAD_VALUE, PAD_VALUE]])

    expected_t = torch.tensor([[0, 1], [1, 2]])

    actual_t = remove_padding(t, PAD_VALUE)

    assert torch.eq(expected_t, actual_t).all()


def test_can_remove_padding_3d():
    t = torch.tensor(
        [[[0, 1], [1, 2]], [[PAD_VALUE, PAD_VALUE], [PAD_VALUE, PAD_VALUE]]]
    )

    expected_t = torch.tensor([[[0, 1], [1, 2]]])

    actual_t = remove_padding(t, PAD_VALUE)

    assert torch.eq(expected_t, actual_t).all()


@pytest.mark.parametrize("shape", [(3, 4), (3, 4, 2), (3, 4, 2, 5)])
def test_can_remove_padding(shape):
    batch_size, *_ = shape

    t = torch.ones(*shape)
    t[-1, ...] = PAD_VALUE

    expected_t = torch.ones(batch_size - 1, *shape[1:])

    actual_t = remove_padding(t, PAD_VALUE)

    assert torch.eq(expected_t, actual_t).all()


@pytest.mark.parametrize(
    "process_decorator",
    [world_process_zero_only, local_process_zero_only, local_process_zero_first],
)
def test_process_decorators_are_transparent_without_multi_gpu(process_decorator):
    @process_decorator
    def fun():
        return 42

    result = fun()

    assert result == 42



# Modifications Copyright © 2022 Chris Hughes
# SchedulerBase adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/scheduler/scheduler.py

from abc import ABC, abstractmethod
from numbers import Number
from typing import Union, Iterable

import torch


class SchedulerBase(ABC):
    """
    A parameter scheduler base class that can be used to update any field within an optimizer's parameter groups.
    The most common use case for this is learning rate scheduling.

    Unlike PyTorch's schedulers, which can be called at different points in the training loop depending on the
    implementation, this class is intended to be consistently called at the end of each optimizer update.

    As this class is stateless by default, it expects that the number of updates is explicitly provided,
    as illustrated below::

        for current_epoch, epoch in enumerate(num_epochs):
            num_updates = current_epoch * num_update_steps_per_epoch
            for batch in train_dataloader:
                xb, yb = batch
                predictions = model(xb)
                loss = loss_func(predictions, yb)

                loss.backward()
                optimizer.step()

                num_updates +=1
                scheduler.step_update(num_updates)
    """

    def __init__(self, optimizer: torch.optim.Optimizer, param_group_field: str = "lr"):
        """
        Create a new instance of a parameter scheduler.

        :param optimizer: a PyTorch optimizer
        :param param_group_field: the field in the optimizer's parameter groups corresponding to the parameter to be scheduled
        """

        self.optimizer = optimizer
        self.param_group_field = param_group_field
        self._initial_param_group_field = f"initial_{param_group_field}"
        self._store_initial_lrs()
        self.base_lr_values = [
            group[self._initial_param_group_field]
            for group in self.optimizer.param_groups
        ]
        self._update_param_groups(self.base_lr_values)

    def _store_initial_lrs(self):
        """
        Store the initial value of the scheduled parameter for each parameter group.
        """
        for i, group in enumerate(self.optimizer.param_groups):
            if self.param_group_field not in group:
                raise KeyError(
                    f"{self.param_group_field} missing from param_groups[{i}]"
                )
            group.setdefault(
                self._initial_param_group_field, group[self.param_group_field]
            )

    @abstractmethod
    def get_updated_values(
        self, num_updates: int
    ) -> Union[None, Number, Iterable[Number]]:
        """
        Calculate updated values for the scheduled parameter.

        If a single value is returned, all parameter groups will be updated with this value.

        To update each parameter group with a different value, an iterable collection,
        containing an updated value for each parameter group, should be returned.

        If None is returned, the parameter groups will not be updated.

        :param num_updates: the number of optimizer updates
        :return: the updated values of the scheduled parameter. This should be either a single value, or an iterable collection containing a value for each parameter group.

        """
        pass

    def step_update(self, num_updates: int):
        """
        Calculate the updated value of the scheduled parameter and update the optimizer's parameter groups.

        :param num_updates: the number of optimizer updates

        """
        values = self.get_updated_values(num_updates)
        if values is not None:
            self._update_param_groups(values)

    def _update_param_groups(self, values):
        """
        Update the scheduled parameter with the given values in all of the optimizer's parameter groups.

        :param values: the updated values of the scheduled parameter
        """
        if not isinstance(values, (list, tuple)):
            values = [values] * len(self.optimizer.param_groups)
        for param_group, value in zip(self.optimizer.param_groups, values):
            param_group[self.param_group_field] = value

    def state_dict(self):
        """
        Get the state dict for the scheduler, containing all attributes except the optimizer,
        which should be saved separately.

        :return: the scheduler's state dict
        """
        return {
            key: value for key, value in self.__dict__.items() if key != "optimizer"
        }

    def load_state_dict(self, state_dict):
        """
        Updates the attributes of the given scheduler from the given state dict.

        :param state_dict: the state dict to be loaded
        """
        self.__dict__.update(state_dict)


class StatefulSchedulerBase(SchedulerBase, ABC):
    """
    A stateful parameter scheduler base class that can be used to update any field within an optimizer's parameter groups.
    The most common use case for this is learning rate scheduling.

    Unlike PyTorch's schedulers, which can be called at different points in the training loop depending on the
    implementation, this class is intended to be consistently called at the end of each optimizer update.

    This class is responsible for maintaining the number of updates, incrementing an internal count each time that
    the scheduler step is calculated.

    The usage of this class is illustrated below::

        for current_epoch, epoch in enumerate(num_epochs):
            for batch in train_dataloader:
                xb, yb = batch
                predictions = model(xb)
                loss = loss_func(predictions, yb)

                loss.backward()
                optimizer.step()

                scheduler.step()
    """

    def __init__(self, optimizer, param_group_field: str = "lr"):
        """
        Create a new instance of a stateful parameter scheduler.

        :param optimizer: a PyTorch optimizer
        :param param_group_field: the field in the optimizer's parameter groups corresponding to the parameter to be scheduled
        """
        super().__init__(optimizer=optimizer, param_group_field=param_group_field)
        self._num_updates = -1

    def step(self):
        """
        Calculate the updated value of the scheduled parameter and update the optimizer's parameter groups.
        """
        self._num_updates += 1
        self.step_update(self._num_updates)



import torch

from pytorch_accelerated.schedulers.scheduler_base import (
    SchedulerBase,
    StatefulSchedulerBase,
)


class EqualIterationScheduler(SchedulerBase):
    def get_updated_values(self, num_updates: int):
        return num_updates


class StatefulEqualIterationScheduler(StatefulSchedulerBase):
    def get_updated_values(self, num_updates: int):
        return num_updates


def create_model_and_optimizer(lr_1, lr_2):
    model = torch.nn.Linear(2, 1)
    params = list(model.parameters())
    optimizer = torch.optim.SGD(
        [
            {
                "params": params[0],
            },
            {"params": params[1], "lr": lr_2},
        ],
        lr=lr_1,
    )
    return model, optimizer


def test_stores_initial_lr_on_creation():
    lr_1 = 0.01
    lr_2 = 0.001
    expected_lrs = [lr_1, lr_2]
    model, optimizer = create_model_and_optimizer(lr_1, lr_2)

    scheduler = EqualIterationScheduler(optimizer)

    for expected_lr, param_group in zip(expected_lrs, optimizer.param_groups):
        assert "initial_lr" in param_group
        assert param_group["initial_lr"] == expected_lr


def test_can_update_parameter_groups():
    lr_1 = 0.01
    lr_2 = 0.001
    expected_initial_lrs = [lr_1, lr_2]
    model, optimizer = create_model_and_optimizer(lr_1, lr_2)

    scheduler = EqualIterationScheduler(optimizer)
    lrs_before_update = [param_group["lr"] for param_group in optimizer.param_groups]
    scheduler.step_update(1)
    lrs_after_update = [param_group["lr"] for param_group in optimizer.param_groups]

    assert lrs_before_update == expected_initial_lrs
    assert lrs_after_update == [1, 1]


def test_stateful_scheduler_can_update_parameter_groups():
    lr_1 = 0.01
    lr_2 = 0.001
    expected_initial_lrs = [lr_1, lr_2]
    model, optimizer = create_model_and_optimizer(lr_1, lr_2)

    scheduler = StatefulEqualIterationScheduler(optimizer)
    lrs_before_update = [param_group["lr"] for param_group in optimizer.param_groups]
    scheduler.step()
    lrs_after_update_1 = [param_group["lr"] for param_group in optimizer.param_groups]
    scheduler.step()
    lrs_after_update_2 = [param_group["lr"] for param_group in optimizer.param_groups]

    assert lrs_before_update == expected_initial_lrs
    assert lrs_after_update_1 == [0, 0]
    assert lrs_after_update_2 == [1, 1]



# Modifications Copyright (C) 2021 Chris Hughes
# Model EMA code adapted from https://github.com/rwightman/pytorch-image-models/blob/master/timm/utils/model_ema.py

import os
from copy import deepcopy
from functools import wraps

import torch
from accelerate.state import PartialState
from accelerate.utils import wait_for_everyone
from torch import nn, Tensor

LIMIT_BATCHES_ENV_VAR = "PT_ACC_LIMIT_BATCHES"


class LimitBatches:
    """
    A context manager which can be used to limit the batches used within a :class:`~pytorch_accelerated.trainer.Trainer`.
    Any Trainer initialised within this context manager will contain the :class:`~pytorch_accelerated.callbacks.LimitBatchesCallback`
    callback. To remove this behaviour, a new trainer must be created or this callback must be explicitly removed.
    """

    def __init__(self, num_batches: int):
        self.num_batches = num_batches

    def __enter__(self):
        os.environ[LIMIT_BATCHES_ENV_VAR] = str(self.num_batches)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        del os.environ[LIMIT_BATCHES_ENV_VAR]


def local_process_zero_only(func):
    """
    A decorator which can be used to ensure that the decorated function is only executed on the local main process
    during distributed training

    :param func: the function to be decorated
    """

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        state = PartialState()
        if state.local_process_index == 0:
            result = func(*args, **kwargs)
            wait_for_everyone()
            return result
        else:
            wait_for_everyone()

    return wrapper_func


def local_process_zero_first(func):
    """
    A decorator which can be used to ensure that the decorated function is executed on the local main process first
    during distributed training

    :param func: the function to be decorated
    """

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        result = None
        state = PartialState()
        if state.local_process_index == 0:
            result = func(*args, **kwargs)
        wait_for_everyone()
        if state.local_process_index != 0:
            result = func(*args, **kwargs)
        return result

    return wrapper_func


def world_process_zero_only(func):
    """
    A decorator which can be used to ensure that the decorated function is only executed on the global main process
    during distributed training

    :param func: the function to be decorated

    """

    @wraps(func)
    def wrapper_func(*args, **kwargs):
        state = PartialState()
        if state.process_index == 0:
            result = func(*args, **kwargs)
            wait_for_everyone()
            return result
        else:
            wait_for_everyone()

    return wrapper_func


class ModelEma(nn.Module):
    """
    Maintains a moving average of everything in the model state_dict (parameters and buffers), based on the ideas
    from https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage.

    This class maintains a copy of the model that we are training. However,
    rather than updating all of the parameters of this model after every update step,
    we set these parameters using a linear combination of the existing parameter values and the updated values

    .. Note:: It is important to note that this class is sensitive to where it is initialised.
        During distributed training, it should be applied before before the conversion to :class:`~torch.nn.SyncBatchNorm`
        takes place and before the :class:`torch.nn.parallel.DistributedDataParallel` wrapper is used!
    """

    def __init__(self, model, decay=0.9999):
        """
        Create an instance of :class:`torch.nn.Module` to maintain an exponential moving average of the weights of
        the given model.

        This is done using the following formula:

        `updated_EMA_model_weights = decay * EMA_model_weights + (1. — decay) * updated_model_weights`

        where the decay is a parameter that we set.

        :param model: the subclass of :class: `torch.nn.Module` that we are training. This is the model that will be updated in our training loop as normal.
        :param decay: the amount of decay to use, which determines how much of the previous state will be maintained. The TensorFlow documentation suggests that reasonable values for decay are close to 1.0, typically in the multiple-nines range: 0.999, 0.9999

        """
        super().__init__()
        # make a copy of the model for accumulating moving average of weights
        self.module = deepcopy(model)
        for p in self.module.parameters():
            p.requires_grad_(False)
        self.module.eval()
        self.decay = decay

    def update_fn(self, ema_model_weights, updated_model_weights):
        return (
            self.decay * ema_model_weights + (1.0 - self.decay) * updated_model_weights
        )

    def _update(self, model, update_fn):
        with torch.no_grad():
            for ema_v, model_v in zip(
                self.module.state_dict().values(), model.state_dict().values()
            ):
                updated_v = update_fn(ema_v, model_v)
                ema_v.copy_(updated_v)

    def update(self, model):
        self._update(model, update_fn=self.update_fn)

    def set(self, model):
        self._update(model, update_fn=lambda e, m: m)


def worker_init_fn(worker_id):
    """
    Function that is used to set the random seed in each dataloader worker.
    This differs from the default by using the current random seed, which should be different in each process,
    rather than the initial random seed.
    """
    return torch.seed() + worker_id


def remove_padding(padded_tensor: Tensor, pad_value):
    """
    A utility function to remove padding tokens from a tensor. This can be useful after applying padding
    in order to gather a tensor.

    :param padded_tensor: the tensor containing padding
    :param pad_value: the padding token to be removed
    :return: a new tensor with padding tokens removed
    """
    padding_mask = padded_tensor == pad_value

    while padding_mask.ndim > 1:
        padding_mask = torch.all(padding_mask, dim=-1)

    result = padded_tensor[~padding_mask, ...]

    return result



import torch

from pytorch_accelerated.schedulers.cosine_scheduler import CosineLrScheduler


def create_model_and_optimizer(lr_1, lr_2):
    model = torch.nn.Linear(2, 1)
    params = list(model.parameters())
    optimizer = torch.optim.SGD(
        [
            {
                "params": params[0],
            },
            {"params": params[1], "lr": lr_2},
        ],
        lr=lr_1,
    )
    return model, optimizer


def collect_lrs_for_scheduler(scheduler, num_epochs, num_steps_per_epoch):
    group_1_lrs = []
    group_2_lrs = []

    for epoch in range(num_epochs):
        for i in range(num_steps_per_epoch):
            scheduler.step()

            group_1_lrs.append(scheduler.optimizer.param_groups[0]["lr"])
            group_2_lrs.append(scheduler.optimizer.param_groups[1]["lr"])

    return group_1_lrs, group_2_lrs


def test_lr_maxes_equal():
    num_epochs = 300
    num_steps_per_epoch = 10
    lr_1_max = 0.01
    lr_2_max = 0.002
    model, optimizer = create_model_and_optimizer(lr_1_max, lr_2_max)

    scheduler = CosineLrScheduler(
        optimizer,
        total_num_epochs=num_epochs,
        num_update_steps_per_epoch=num_steps_per_epoch,
    )
    group_1_lrs, group_2_lrs = collect_lrs_for_scheduler(
        scheduler, num_epochs, num_steps_per_epoch
    )

    assert group_1_lrs[0] == lr_1_max
    assert max(group_1_lrs) == lr_1_max
    assert group_2_lrs[0] == lr_2_max
    assert max(group_2_lrs) == lr_2_max


def test_lr_mins_at_schedule_end():
    num_epochs = 1
    num_steps_per_epoch = 10
    lr_1_max = 0.01
    lr_2_max = 0.002
    lr_min = 0.0001
    model, optimizer = create_model_and_optimizer(lr_1_max, lr_2_max)

    scheduler = CosineLrScheduler(
        optimizer,
        total_num_epochs=num_epochs,
        num_update_steps_per_epoch=num_steps_per_epoch,
        lr_min=lr_min,
    )
    group_1_lrs, group_2_lrs = collect_lrs_for_scheduler(
        scheduler, num_epochs, num_steps_per_epoch + 1
    )

    assert group_1_lrs[-1] == lr_min
    assert min(group_1_lrs) == lr_min
    assert group_2_lrs[-1] == lr_min
    assert min(group_2_lrs) == lr_min


def test_lr_min_ratio_at_schedule_end():
    num_epochs = 1
    num_steps_per_epoch = 10
    lr_1_max = 0.01
    lr_2_max = 0.002
    lr_min_ratio = 0.5
    expected_lr_1_min = lr_min_ratio * lr_1_max
    expected_lr_2_min = lr_min_ratio * lr_2_max
    model, optimizer = create_model_and_optimizer(lr_1_max, lr_2_max)

    scheduler = CosineLrScheduler(
        optimizer,
        total_num_epochs=num_epochs,
        num_update_steps_per_epoch=num_steps_per_epoch,
        min_lr_ratio=lr_min_ratio,
    )
    group_1_lrs, group_2_lrs = collect_lrs_for_scheduler(
        scheduler, num_epochs, num_steps_per_epoch + 1
    )

    assert group_1_lrs[-1] == expected_lr_1_min
    assert min(group_1_lrs) == expected_lr_1_min
    assert group_2_lrs[-1] == expected_lr_2_min
    assert min(group_2_lrs) == expected_lr_2_min


def test_cooldown_epochs_at_lr_min():
    num_epochs = 10
    num_cooldown_epochs = 2
    num_steps_per_epoch = 10
    num_cooldown_steps = num_cooldown_epochs * num_steps_per_epoch
    lr_1_max = 0.01
    lr_2_max = 0.002
    lr_min = 1e-6
    expected_cooldown_steps = [lr_min] * num_cooldown_steps
    model, optimizer = create_model_and_optimizer(lr_1_max, lr_2_max)

    scheduler = CosineLrScheduler(
        optimizer,
        total_num_epochs=num_epochs,
        num_update_steps_per_epoch=num_steps_per_epoch,
        lr_min=lr_min,
        num_cooldown_epochs=num_cooldown_epochs,
    )
    group_1_lrs, group_2_lrs = collect_lrs_for_scheduler(
        scheduler, num_epochs, num_steps_per_epoch
    )

    assert group_1_lrs[-num_cooldown_steps:] == expected_cooldown_steps
    assert group_2_lrs[-num_cooldown_steps:] == expected_cooldown_steps


def test_warmup():
    num_epochs = 10
    num_warmup_epochs = 2
    num_steps_per_epoch = 10
    num_warmup_steps = num_warmup_epochs * num_steps_per_epoch
    lr_1_max = 0.01
    lr_2_max = 0.002
    warmup_lr = 1e-6
    model, optimizer = create_model_and_optimizer(lr_1_max, lr_2_max)

    scheduler = CosineLrScheduler(
        optimizer,
        total_num_epochs=num_epochs,
        num_update_steps_per_epoch=num_steps_per_epoch,
        num_warmup_epochs=num_warmup_epochs,
        warmup_starting_lr=warmup_lr,
    )
    group_1_lrs, group_2_lrs = collect_lrs_for_scheduler(
        scheduler, num_epochs, num_steps_per_epoch
    )

    assert group_1_lrs[0] == warmup_lr
    assert group_2_lrs[0] == warmup_lr
    assert group_1_lrs[num_warmup_steps] == max(group_1_lrs)
    assert group_2_lrs[num_warmup_steps] == max(group_2_lrs)


def test_warmup_from_ratio():
    num_epochs = 10
    num_warmup_epochs = 2
    num_steps_per_epoch = 10
    num_warmup_steps = num_warmup_epochs * num_steps_per_epoch
    lr_1_max = 0.01
    lr_2_max = 0.002
    warmup_lr_ratio = 0.01
    starting_lr_1 = warmup_lr_ratio * lr_1_max
    starting_lr_2 = warmup_lr_ratio * lr_2_max
    model, optimizer = create_model_and_optimizer(lr_1_max, lr_2_max)

    scheduler = CosineLrScheduler(
        optimizer,
        total_num_epochs=num_epochs,
        num_update_steps_per_epoch=num_steps_per_epoch,
        num_warmup_epochs=num_warmup_epochs,
        warmup_starting_lr_ratio=warmup_lr_ratio,
    )
    group_1_lrs, group_2_lrs = collect_lrs_for_scheduler(
        scheduler, num_epochs, num_steps_per_epoch
    )

    assert group_1_lrs[0] == starting_lr_1
    assert group_2_lrs[0] == starting_lr_2
    assert group_1_lrs[num_warmup_steps] == max(group_1_lrs)
    assert group_2_lrs[num_warmup_steps] == max(group_2_lrs)


def test_warmup_and_cooldown():
    num_epochs = 10
    num_warmup_epochs = 2
    num_cooldown_epochs = 2
    num_steps_per_epoch = 10
    num_warmup_steps = num_warmup_epochs * num_steps_per_epoch
    num_cooldown_steps = num_cooldown_epochs * num_steps_per_epoch
    lr_1_max = 0.01
    lr_2_max = 0.002
    lr_min = 1e-6
    expected_cooldown_steps = [lr_min] * num_cooldown_steps
    model, optimizer = create_model_and_optimizer(lr_1_max, lr_2_max)

    scheduler = CosineLrScheduler(
        optimizer,
        total_num_epochs=num_epochs,
        num_update_steps_per_epoch=num_steps_per_epoch,
        num_warmup_epochs=num_warmup_epochs,
        warmup_starting_lr=lr_min,
        lr_min=lr_min,
        num_cooldown_epochs=2,
    )
    group_1_lrs, group_2_lrs = collect_lrs_for_scheduler(
        scheduler, num_epochs, num_steps_per_epoch
    )

    assert group_1_lrs[0] == lr_min
    assert group_2_lrs[0] == lr_min
    assert group_1_lrs[num_warmup_steps] == max(group_1_lrs)
    assert group_2_lrs[num_warmup_steps] == max(group_2_lrs)
    assert group_1_lrs[-num_cooldown_steps:] == expected_cooldown_steps
    assert group_2_lrs[-num_cooldown_steps:] == expected_cooldown_steps



import os
import random
from pathlib import Path

import numpy as np
import pandas as pd
import torch
from evaluation.calculate_map_callback import CalculateMeanAveragePrecisionCallback
from frcnn.dataset import FasterRCNNDataset, get_eval_transforms, get_train_transforms
from frcnn.model import create_frcnn_model
from frcnn.trainer import FasterRCNNTrainer, faster_rcnn_collate_fn
from func_to_script import script
from PIL import Image
from torch.utils.data import Dataset

from pytorch_accelerated.callbacks import get_default_callbacks
from pytorch_accelerated.schedulers import CosineLrScheduler


def load_cars_df(annotations_file_path, images_path):
    all_images = sorted(set([p.parts[-1] for p in images_path.iterdir()]))
    image_id_to_image = {i: im for i, im in enumerate(all_images)}
    image_to_image_id = {v: k for k, v, in image_id_to_image.items()}

    annotations_df = pd.read_csv(annotations_file_path)
    annotations_df.loc[:, "class_name"] = "car"
    annotations_df.loc[:, "has_annotation"] = True

    # add 100 empty images to the dataset
    empty_images = sorted(set(all_images) - set(annotations_df.image.unique()))
    non_annotated_df = pd.DataFrame(list(empty_images)[:100], columns=["image"])
    non_annotated_df.loc[:, "has_annotation"] = False
    non_annotated_df.loc[:, "class_name"] = "background"

    df = pd.concat((annotations_df, non_annotated_df))

    # As 0 is the background class, class ids must start at 1
    class_id_to_label = {
        i + 1: c
        for i, c in enumerate(df.query("has_annotation == True").class_name.unique())
    }
    class_label_to_id = {v: k for k, v in class_id_to_label.items()}

    df["image_id"] = df.image.map(image_to_image_id)
    df["class_id"] = df.class_name.map(class_label_to_id)

    file_names = tuple(df.image.unique())
    random.seed(42)
    validation_files = set(random.sample(file_names, int(len(df) * 0.2)))
    train_df = df[~df.image.isin(validation_files)]
    valid_df = df[df.image.isin(validation_files)]

    lookups = {
        "image_id_to_image": image_id_to_image,
        "image_to_image_id": image_to_image_id,
        "class_id_to_label": class_id_to_label,
        "class_label_to_id": class_label_to_id,
    }
    return train_df, valid_df, lookups


class DatasetAdaptor(Dataset):
    def __init__(
        self,
        images_dir_path,
        annotations_dataframe,
        transforms=None,
    ):
        self.images_dir_path = Path(images_dir_path)
        self.annotations_df = annotations_dataframe
        self.transforms = transforms

        self.image_idx_to_image_id = {
            idx: image_id
            for idx, image_id in enumerate(self.annotations_df.image_id.unique())
        }
        self.image_id_to_image_idx = {
            v: k for k, v, in self.image_idx_to_image_id.items()
        }

    def __len__(self) -> int:
        return len(self.image_idx_to_image_id)

    def __getitem__(self, index):
        image_id = self.image_idx_to_image_id[index]
        image_info = self.annotations_df[self.annotations_df.image_id == image_id]
        file_name = image_info.image.values[0]
        assert image_id == image_info.image_id.values[0]

        image = Image.open(self.images_dir_path / file_name).convert("RGB")
        image = np.array(image)

        if image_info.has_annotation.any():
            xyxy_bboxes = image_info[["xmin", "ymin", "xmax", "ymax"]].values
            class_ids = image_info["class_id"].values
        else:
            xyxy_bboxes = np.array([])
            class_ids = np.array([])

        if self.transforms is not None:
            transformed = self.transforms(
                image=image, bboxes=xyxy_bboxes, labels=class_ids
            )
            image = transformed["image"]
            xyxy_bboxes = np.array(transformed["bboxes"])
            class_ids = np.array(transformed["labels"])

        return image, xyxy_bboxes, class_ids, image_id


DATA_PATH = Path("/".join(Path(__file__).absolute().parts[:-3])) / "data/cars"


@script
def main(
    data_path: str = DATA_PATH,
    image_size: int = 800,
    num_epochs: int = 30,
    batch_size: int = 8,
):
    data_path = Path(data_path)
    images_path = data_path / "training_images"
    annotations_file_path = data_path / "annotations.csv"

    train_df, valid_df, lookups = load_cars_df(annotations_file_path, images_path)

    train_ds = DatasetAdaptor(
        images_path,
        train_df,
    )
    eval_ds = DatasetAdaptor(images_path, valid_df)

    train_dataset = FasterRCNNDataset(train_ds, transforms=get_train_transforms())
    eval_dataset = FasterRCNNDataset(eval_ds, transforms=get_eval_transforms())

    num_classes = 1

    model = create_frcnn_model(num_classes=num_classes, image_size=image_size)
    optimizer = torch.optim.AdamW(model.parameters(), lr=0.0001)

    trainer = FasterRCNNTrainer(
        model=model,
        optimizer=optimizer,
        callbacks=[
            CalculateMeanAveragePrecisionCallback.create_from_targets_df(
                targets_df=valid_df.query("has_annotation == True")[
                    ["image_id", "xmin", "ymin", "xmax", "ymax", "class_id"]
                ],
                image_ids=set(valid_df.image_id.unique()),
                iou_threshold=0.2,
            ),
            *get_default_callbacks(progress_bar=True),
        ],
    )

    trainer.train(
        num_epochs=num_epochs,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        per_device_batch_size=batch_size,
        create_scheduler_fn=CosineLrScheduler.create_scheduler_fn(
            num_warmup_epochs=5,
            num_cooldown_epochs=5,
        ),
        collate_fn=faster_rcnn_collate_fn,
    )


if __name__ == "__main__":
    main()



from pathlib import Path
from tempfile import TemporaryFile, TemporaryDirectory
from unittest.mock import MagicMock, Mock, call
import pytest

import torch
from pytest import fixture
from pytorch_accelerated.utils import worker_init_fn
from torch import optim, nn

from pytorch_accelerated.trainer import Trainer


class DummyTrainer(Trainer):
    def __init__(
        self,
        train_dl_mock=MagicMock(),
        eval_dl_mock=MagicMock,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.train_dl_mock = train_dl_mock
        self.eval_dl_mock = eval_dl_mock

    def calculate_train_batch_loss(self, batch):
        return {
            "loss": torch.tensor(1.0),
            "model_outputs": MagicMock(),
            "batch_size": 1,
        }

    def create_train_dataloader(self, batch_size, train_dl_kwargs):
        return self.train_dl_mock

    def create_eval_dataloader(self, batch_size, eval_dl_kwargs):
        return self.eval_dl_mock

    def calculate_eval_batch_loss(self, batch):
        return {
            "loss": torch.tensor(1.0),
            "model_outputs": MagicMock(),
            "batch_size": 1,
        }

    def _prepare_model_and_optimizer(self):
        pass


class SimpleModel(nn.Module):
    def __init__(self, input_size, output_size):
        super().__init__()
        self.linear = nn.Linear(input_size, output_size)

    def forward(self, x):
        return self.linear(x)


@fixture
def model():
    return SimpleModel(10, 1)


@fixture
def optimizer(model):
    return optim.SGD(model.parameters(), lr=0.01)


def test_skip_eval_if_not_present(mocker):
    trainer = DummyTrainer(model=Mock(), optimizer=Mock(), loss_func=Mock())
    mocked_train_epoch: MagicMock = mocker.patch.object(trainer, "_run_train_epoch")
    mocked_eval_epoch = mocker.patch.object(trainer, "_run_eval_epoch")

    trainer.train(train_dataset=Mock(), num_epochs=2)

    mocked_train_epoch.assert_has_calls(
        [call(trainer._train_dataloader), call(trainer._train_dataloader)]
    )
    mocked_eval_epoch.assert_not_called()


def test_skip_scheduler_step_if_not_present(mocker):
    trainer = DummyTrainer(model=Mock(), optimizer=Mock(), loss_func=Mock())
    scheduler_step = mocker.patch.object(trainer, "scheduler_step")

    trainer.train(train_dataset=Mock(), num_epochs=1)

    scheduler_step.assert_not_called()


def test_can_override_train_dataloader_kwargs(mocker):
    trainer = Trainer(model=Mock, optimizer=Mock(), loss_func=Mock())
    dl_constructor: MagicMock = mocker.patch("pytorch_accelerated.trainer.DataLoader")
    train_dataset = Mock()
    collate_fn = Mock()
    trainer.train_dataset = train_dataset
    trainer.collate_fn = collate_fn
    override_dl_kwargs = {"batch_size": 100, "pin_memory": False, "num_workers": 0}
    expected_dl_kwargs = {
        "shuffle": True,
        "batch_size": 100,
        "pin_memory": False,
        "num_workers": 0,
        "worker_init_fn": worker_init_fn,
    }

    trainer.create_train_dataloader(10, override_dl_kwargs)

    dl_constructor.assert_called_with(
        dataset=train_dataset, collate_fn=collate_fn, **expected_dl_kwargs
    )


def test_can_override_eval_dataloader_kwargs(mocker):
    trainer = Trainer(model=Mock, optimizer=Mock(), loss_func=Mock())
    dl_constructor: MagicMock = mocker.patch("pytorch_accelerated.trainer.DataLoader")
    eval_dataset = Mock()
    collate_fn = Mock()
    trainer.eval_dataset = eval_dataset
    trainer.collate_fn = collate_fn
    override_dl_kwargs = {"batch_size": 100, "pin_memory": False, "num_workers": 0}
    expected_dl_kwargs = {
        "shuffle": False,
        "batch_size": 100,
        "pin_memory": False,
        "num_workers": 0,
        "worker_init_fn": worker_init_fn,
    }

    trainer.create_eval_dataloader(10, override_dl_kwargs)

    dl_constructor.assert_called_with(
        dataset=eval_dataset, collate_fn=collate_fn, **expected_dl_kwargs
    )


def test_model_is_in_correct_mode():
    pass


def test_gradient_accumulation():
    pass


def test_can_create_scheduler():
    pass


def test_can_reset_run_history():
    pass


def test_check_eval_batch_size_is_transparent_on_single_process():
    batch_size = 8
    n_samples = batch_size - 1

    class FakeRunConfig:
        eval_total_batch_size = batch_size
        is_distributed = False

    trainer = Trainer("irrelevant", "irrelevant", "irrelevant")
    trainer.eval_dataset = list(range(n_samples))
    trainer.run_config = FakeRunConfig()

    trainer._check_eval_batch_size()


def test_check_eval_batch_size_raises_batch_size_bigger_than_dataset():
    batch_size = 8

    class FakeRunConfig:
        eval_total_batch_size = batch_size
        is_distributed = True

    trainer = Trainer("irrelevant", "irrelevant", "irrelevant")
    trainer.eval_dataset = list(range(batch_size - 1))
    trainer.run_config = FakeRunConfig()

    with pytest.raises(ValueError):
        trainer._check_eval_batch_size()


def test_check_eval_batch_size_raises_empty_node_on_last_batch():
    per_device_batch_size = 8
    n_processes = 4
    n_full_batches = 2

    n_samples_full_batches = per_device_batch_size * n_processes * n_full_batches
    # One process will have no samples, another one less samples than the batch size
    n_samples_last_batch = per_device_batch_size * (n_processes - 1) - 1
    n_samples = n_samples_full_batches + n_samples_last_batch

    class FakeRunConfig:
        eval_per_device_batch_size = per_device_batch_size
        eval_total_batch_size = per_device_batch_size * n_processes
        num_processes = n_processes
        is_distributed = True

    trainer = Trainer("irrelevant", "irrelevant", "irrelevant")
    trainer.eval_dataset = list(range(n_samples))
    trainer.run_config = FakeRunConfig()

    with pytest.warns(UserWarning):
        trainer._check_eval_batch_size()


def test_check_eval_batch_size_warns_padding_is_needed():
    per_device_batch_size = 8
    n_processes = 4
    n_full_batches = 2

    n_samples_full_batches = per_device_batch_size * n_processes * n_full_batches
    # One process will have just one sample
    n_samples_last_batch = per_device_batch_size * (n_processes - 1) + 1
    n_samples = n_samples_full_batches + n_samples_last_batch

    class FakeRunConfig:
        eval_per_device_batch_size = per_device_batch_size
        eval_total_batch_size = per_device_batch_size * n_processes
        num_processes = n_processes
        is_distributed = True

    trainer = Trainer("irrelevant", "irrelevant", "irrelevant")
    trainer.eval_dataset = list(range(n_samples))
    trainer.run_config = FakeRunConfig()

    with pytest.warns(UserWarning):
        trainer._check_eval_batch_size()
